import unittest


class ValidatorTests(unittest.TestCase):
    """
    Test the validator class

    """
    def setUp(self):
        """
        Setup validator
        """
        pass

    def tearDown(self):
        """
        Delete validator
        """
        pass

